local tilescale = 64
local minechance = 10
local renderdist = 4
local airacivate = false
local iframeabuse = 1
local negativemines = false
local cachepertic = 4

local flagplacex = 0
local flagplacey = 0

if not savedtiles then
	rawset(_G, "savedtiles", {})
end

if not flagedtiles then
	rawset(_G, "flagedtiles", {})
end

if not cachedtiles then
	rawset(_G, "cachedtiles", {})
end

freeslot("SPR_GRBD")

addHook("MapChange", function(cmap)
	savedtiles = {}
	flagedtiles = {}
	cachedtiles = {}
end)

addHook("NetVars", function(host)
	savedtiles = host($)
	flagedtiles = host($)
	cachedtiles = host($)
	tilescale = host($)
	minechance = host($)
	renderdist = host($)
	airacivate = host($)
	iframeabuse = host($)
	negativemines = host($)
	cachepertic = host($)
end)

local tiletonumb = {
	[22] = -8,
	[21] = -7,
	[20] = -6,
	[19] = -5,
	[18] = -4,
	[17] = -3,
	[16] = -2,
	[15] = -1,
	[14] = 0,
	[2] = 1,
	[3] = 2,
	[4] = 3,
	[5] = 4,
	[6] = 5,
	[7] = 6,
	[8] = 7,
	[9] = 8
}

local function IsPozValid(x,y)
	local pos = P_CeilingzAtPos((x*tilescale*FRACUNIT)+(tilescale*4*FRACUNIT/10), (y*tilescale*FRACUNIT)+(tilescale*FRACUNIT/10), 0, 0) - P_FloorzAtPos((x*tilescale*FRACUNIT)+(tilescale*4*FRACUNIT/10), (y*tilescale*FRACUNIT)+(tilescale*4*FRACUNIT/10), 0, 0)
	local poz = P_CeilingzAtPos((x*tilescale*FRACUNIT)+(tilescale*6*FRACUNIT/10), (y*tilescale*FRACUNIT)+(tilescale*6*FRACUNIT/10), 0, 0) - P_FloorzAtPos((x*tilescale*FRACUNIT)+(tilescale*6*FRACUNIT/10), (y*tilescale*FRACUNIT)+(tilescale*6*FRACUNIT/10), 0, 0)
	return ((R_PointInSubsectorOrNil((x*tilescale*FRACUNIT)+(tilescale*4*FRACUNIT/10), (y*tilescale*FRACUNIT)+(tilescale*4*FRACUNIT/10)) ~= nil or R_PointInSubsectorOrNil((x*tilescale*FRACUNIT)+(tilescale*6*FRACUNIT/10), (y*tilescale*FRACUNIT)+(tilescale*6*FRACUNIT/10)) ~= nil) and pos + poz > 0)
end

local function GetTilePos(xpos,ypos)
	if xpos ~= nil and ypos ~= nil then
		local loopint = FRACUNIT/tilescale
		local pos = (xpos%loopint) + (ypos%loopint)*loopint
		return pos
	end
end

local function GetIndexToPos(ind)
	if ind ~= nil then
		local loopint = FRACUNIT/tilescale
		local x = ind%loopint
		local y = (ind/loopint)%loopint
		return {x,y}
	end
end

local function GenerateTile(ind, chance)
	if savedtiles[ind] then return end
	if not chance then chance = minechance end
	if P_RandomChance(chance*FRACUNIT/100) then
		savedtiles[ind] = -2
	elseif negativemines and P_RandomChance(chance*FRACUNIT/100) then
		savedtiles[ind] = -3
	else
		savedtiles[ind] = -1
	end
end

local function GenerateTileRadius(x, y, r)
	local d = i*2+1
	for i = 1,d^2 do
		local xoff = (i%d)-r
		local yoff = ((i/d)%d)-r
		local poz = GetTilePos(x+xoff,y+yoff)
		if not savedtiles[poz] then
			if not IsPozValid(x+xoff,y+yoff) then 
				savedtiles[poz] = -1 
				continue
			end
			GenerateTile(poz, minechance)
		end
	end
end

local function CheckNeighbors(x,y)
	GenerateTileRadius(x, y, 1)
	local val = 0
	local minefound = false
	for c = 1,9 do
		local xoff = (c%3)-1
		local yoff = ((c/3)%3)-1
		if yoff == 0 and xoff == 0 then continue end
		if not IsPozValid(x+xoff,y+yoff) then continue end
		local loloff = GetTilePos(x+xoff,y+yoff)
		if savedtiles[loloff] and (savedtiles[loloff] == -2 or savedtiles[loloff] == 10) then
			val = $ + 1
			minefound = true
		end
		if savedtiles[loloff] and (savedtiles[loloff] == -3 or savedtiles[loloff] == 11) then
			val = $ - 1
			minefound = true
		end
	end
	if not minefound then
		savedtiles[GetTilePos(x,y)] = 1
		table.insert(cachedtiles, {x,y})
	elseif val < 1 and minefound
		savedtiles[GetTilePos(x,y)] = 14 - val
	elseif minefound and val >= 1
		savedtiles[GetTilePos(x,y)] = val + 1
	end
	return savedtiles[GetTilePos(x,y)]
end

local function QuickSweepTiles(x,y,mo)
	GenerateTileRadius(x, y, 1)
	if savedtiles[GetTilePos(x,y)] == -1 or savedtiles[GetTilePos(x,y)] == -2 or savedtiles[GetTilePos(x,y)] == -3 then
		local loloff = GetTilePos(x,y)
		if flagedtiles[loloff] then return end
		if savedtiles[loloff] == -1 then 
			CheckNeighbors(x,y)
		elseif savedtiles[loloff] == -2 then 
			savedtiles[loloff] = 10
			if mo and mo.valid then
				local lmao = P_SpawnMobj(tilescale*x*FRACUNIT+((tilescale/2)*FRACUNIT),tilescale*y*FRACUNIT+((tilescale/2)*FRACUNIT),mo.z, MT_TNTBARREL)
				lmao.scale = tilescale*FRACUNIT/48
				P_DamageMobj(lmao, mo, mo)
				P_DamageMobj(mo,lmao,lmao)
			end
		else 
			savedtiles[loloff] = 11 
			if mo and mo.valid then
				local lmao = P_SpawnMobj(tilescale*x*FRACUNIT+((tilescale/2)*FRACUNIT),tilescale*y*FRACUNIT+((tilescale/2)*FRACUNIT),mo.z, MT_TNTBARREL)
				lmao.scale = tilescale*FRACUNIT/48
				P_DamageMobj(lmao, mo, mo)
				P_DamageMobj(mo)
			end
		end
		return
	end
	local flags = 0
	local sgalf = 0
	local flagfound = false
	if not tiletonumb[savedtiles[GetTilePos(x,y)]] then
		return
	end
	for c = 1,9 do
		local xoff = (c%3)-1
		local yoff = ((c/3)%3)-1
		if not IsPozValid(x+xoff,y+yoff) then continue end
		if yoff == 0 and xoff == 0 then continue end
		local loloff = GetTilePos(x+xoff,y+yoff)
		if tiletonumb[savedtiles[loloff]] ~= nil then continue end
		if (flagedtiles[loloff] and flagedtiles[loloff] == 1) or savedtiles[loloff] == 10 then
			flags = $ + 1
			flagfound = true
		elseif (flagedtiles[loloff] and flagedtiles[loloff] == 2) or savedtiles[loloff] == 11 then
			sgalf = $ + 1
			flagfound = true
		end
	end
	if tiletonumb[savedtiles[GetTilePos(x,y)]] == flags - sgalf then
		for c = 1,9 do
			local xoff = (c%3)-1
			local yoff = ((c/3)%3)-1
			if not IsPozValid(x+xoff,y+yoff) then continue end
			if yoff == 0 and xoff == 0 then continue end
			local loloff = GetTilePos(x+xoff,y+yoff)
			if not flagedtiles[loloff] then
				if (savedtiles[loloff] == -2 or savedtiles[loloff] == -3) and mo and mo.valid then
					if savedtiles[loloff] == -2 then savedtiles[loloff] = 10 else savedtiles[loloff] = 11 end
					local lmao = P_SpawnMobj(tilescale*x*FRACUNIT+((tilescale/2)*FRACUNIT),tilescale*y*FRACUNIT+((tilescale/2)*FRACUNIT),mo.z, MT_TNTBARREL)
					lmao.scale = tilescale*FRACUNIT/48
					P_DamageMobj(lmao, mo, mo)
					P_DamageMobj(mo)
				elseif savedtiles[loloff] < 0 then
					CheckNeighbors(x+xoff,y+yoff)
				end
			end
		end
	end
end

local function CheckNeighborsRadius(x,y,r)
	local d = i*2+1
	for i = 1,d^2
		local xoff = (i%d)-r
		local yoff = ((i/d)%d)-r
		if not IsPozValid(x+xoff,y+yoff) then continue end
		local poz = GetTilePos(x+xoff,y+yoff)
		if savedtiles[poz] and savedtiles[poz] == -1 then
			CheckNeighbors(x+xoff,y+yoff)
		end
	end
end

addHook("PlayerSpawn", function(player)
	if leveltime == 0 then
		local x = FixedInt(FixedFloor(player.mo.x/tilescale))
		local y = FixedInt(FixedFloor(player.mo.y/tilescale))
		for c = 1,9 do
			local xoff = (c%3)-1
			local yoff = ((c/3)%3)-1
			local indpos = GetTilePos(x+xoff,y+yoff)
			savedtiles[indpos] = -1
		end
	end
end)

addHook("PlayerThink", function(player)
	if player.mo then
		local x = FixedInt(FixedFloor(player.mo.x/tilescale))
		local y = FixedInt(FixedFloor(player.mo.y/tilescale))
		local indpos = GetTilePos(x,y)
		if not (player.cmd.buttons & BT_TOSSFLAG) then
			player.flagplace = false
			player.lockang = player.mo.angle
			player.lockaim = player.aiming
			if player == consoleplayer then
				flagplacex = 0
				flagplacey = 0
			end
		else
			player.flagplace = true
			player.powers[pw_nocontrol] = max($, 2)
			if player.lockang then
				player.mo.angle = player.lockang
			else
				player.lockang = player.mo.angle
			end
			if player.lockaim then
				player.aiming = player.lockaim
			else
				player.lockaim = player.aiming
			end
			if player == consoleplayer then
				flagplacex = max(min($ + mouse.rdx + input.joyAxis(JA_TURN), 272),-272)
				flagplacey = max(min($ + mouse.rdy + input.joyAxis(JA_LOOK), 272),-272)
				local flagx = FixedInt(FixedFloor(FRACUNIT/2+FixedDiv(flagplacex*FRACUNIT, FixedMul(545*FRACUNIT, FRACUNIT/17))))
				local flagy = FixedInt(FixedFloor(FRACUNIT/2+FixedDiv(flagplacey*FRACUNIT, FixedMul(545*FRACUNIT, FRACUNIT/17))))
				if ((player.cmd.buttons & BT_SPIN) and not (player.lastbuttons & BT_SPIN)) or ((player.cmd.buttons & BT_FIRENORMAL) and not (player.lastbuttons & BT_FIRENORMAL)) then
					COM_BufInsertText(player, "setflag "+flagx+" "+flagy)
				end
				if ((player.cmd.buttons & BT_JUMP) and not (player.lastbuttons & BT_JUMP)) or ((player.cmd.buttons & BT_ATTACK) and not (player.lastbuttons & BT_ATTACK)) then
					COM_BufInsertText(player, "quicksweep "+flagx+" "+flagy)
				end
			end
		end
		
		if (P_IsObjectOnGround(player.mo) or (player.mo.eflags & MFE_JUSTHITFLOOR) or airacivate) and indpos then
			GenerateTileRadius(x, y, 1)
			flagedtiles[indpos] = nil
			if savedtiles[indpos] == -2 or savedtiles[indpos] == -3 then
				if savedtiles[indpos] == -2 then
					savedtiles[indpos] = 10
				else
					savedtiles[indpos] = 11
				end
				if (player.powers[pw_flashing] > 0 and iframeabuse == 2) or (player.powers[pw_flashing] <= TICRATE*2 and iframeabuse == 1) then
					player.powers[pw_flashing] = 0
				end
				local lmao = P_SpawnMobj(tilescale*x*FRACUNIT+((tilescale/2)*FRACUNIT),tilescale*y*FRACUNIT+((tilescale/2)*FRACUNIT),player.mo.z, MT_TNTBARREL)
				lmao.scale = tilescale*FRACUNIT/48
				P_DamageMobj(lmao, player.mo, player.mo)
			elseif savedtiles[indpos] == -1 then
				if CheckNeighbors(x,y) == 1 then
					table.insert(cachedtiles, {x,y})
				end
			end
		end
	end
end)

hud.add( function(v, player, camera)
	if player.mo then
		local scale = min(v.dupy(), 2)
		local x = FixedInt(FixedFloor(player.mo.x/tilescale))
		local y = FixedInt(FixedFloor(player.mo.y/tilescale))
		local hudr = 4*scale
		local hudd = hudr*2+1
		local grid = v.cachePatch("MISWGRID")
		v.draw(320-(hudd+1)*6, 0, grid, V_SNAPTORIGHT|V_SNAPTOTOP)
		for i = 1,hudd^2 do
			local xoff = (i%hudd)-hudr
			local yoff = ((i/hudd)%hudd)-hudr
			local pos = GetTilePos(x+xoff,y+yoff)
			local sprite = 0
			if not IsPozValid(x+xoff,y+yoff) then 
				sprite = v.getSpritePatch(SPR_GRBD, L, 0, 0)
				v.drawScaled(((320 - hudr*6) - (yoff+1)*6)*FRACUNIT, ((5 + hudd*3) - xoff*6)*FRACUNIT, FRACUNIT/(2*scale), sprite, V_SNAPTOTOP|V_SNAPTORIGHT)
			else
				if not savedtiles[pos] then
					local frame = A
					if flagedtiles[pos]
						frame = C
					else
						frame = A
					end
					sprite = v.cachePatch("HUDTILE")
				elseif savedtiles[pos] and savedtiles[pos] ~= 1 then
					local frame = A
					if savedtiles[pos] <= -1 then
						if flagedtiles[pos]
							if flagedtiles[pos] == 1 then
								frame = C
							else
								frame = N
							end
						else
							frame = A
						end
					elseif savedtiles[pos] == 10 then
						frame = B
					elseif savedtiles[pos] == 11 then
						frame = M
					elseif savedtiles[pos] < 10
						frame = M - savedtiles[pos]
					else
						frame = savedtiles[pos]
					end
					if frame == A then
						sprite = v.cachePatch("HUDTILE")
					else
						sprite = v.getSpritePatch(SPR_GRBD, frame, 0, 0)
					end
				end
				if not (savedtiles[pos] and savedtiles[pos] == 1) then
					v.drawScaled(((320 - hudr*6) - (yoff+1)*6)*FRACUNIT, ((5 + hudd*3) - xoff*6)*FRACUNIT, FRACUNIT/(2*scale), sprite, V_SNAPTOTOP|V_SNAPTORIGHT)
				end
			end
		end
		local playsprite = v.getSprite2Patch(player.realmo.skin, "XTRA", false, C, 0)
		v.drawScaled((320 - (hudr+1)*6)*FRACUNIT, (7+hudr*6)*FRACUNIT, FRACUNIT/(2*scale), playsprite, V_SNAPTOTOP|V_SNAPTORIGHT, v.getColormap(player.realmo.skin, player.skincolor))
		if player.flagplace and player == consoleplayer then
			local cursor = v.cachePatch("MMOUSE")
			local flag = v.getSpritePatch(SPR_GRBD, C, 0, 0)
			local flagx = FixedInt(FixedFloor(FRACUNIT/2+FixedDiv(flagplacex*FRACUNIT, FixedMul(545*FRACUNIT, FRACUNIT/17))))
			local flagy = FixedInt(FixedFloor(FRACUNIT/2+FixedDiv(flagplacey*FRACUNIT, FixedMul(545*FRACUNIT, FRACUNIT/17))))
			v.drawScaled(((320 - hudr*6) + (flagx-1)*6)*FRACUNIT, ((5 + hudd*3) + flagy*6)*FRACUNIT, FRACUNIT/(2*scale), flag, V_SNAPTOTOP|V_SNAPTORIGHT|V_TRANSLUCENT)
			v.drawScaled((320 - (hudr+1)*6)*FRACUNIT + flagplacex*12288, (3 + hudd*3)*FRACUNIT + flagplacey*12288, FRACUNIT/2, cursor, V_SNAPTORIGHT|V_SNAPTOTOP)
			
			v.drawString(318 - hudr*7, 20 + (hudd+1)*6, "Spin/NormalFire to place/remove flags", V_SNAPTORIGHT|V_SNAPTOTOP|V_ALLOWLOWERCASE, "small-thin-right")
			v.drawString(318 - hudr*7, 25 + (hudd+1)*6, "Jump/Fire to Check Tiles", V_SNAPTORIGHT|V_SNAPTOTOP|V_ALLOWLOWERCASE, "small-thin-right")
			v.drawString(318, 2 + (hudd+1)*6, (x-flagx)+" mX", V_SNAPTORIGHT|V_SNAPTOTOP, "right")
			v.drawString(318, 10 + (hudd+1)*6, (y-flagy)+" mY", V_SNAPTORIGHT|V_SNAPTOTOP, "right")
		else
			v.drawString(318, 2 + (hudd+1)*6, x+" X", V_SNAPTORIGHT|V_SNAPTOTOP, "right")
			v.drawString(318, 10 + (hudd+1)*6, y+" Y", V_SNAPTORIGHT|V_SNAPTOTOP, "right")
		end
	end
end)

addHook("ThinkFrame", function()
	if #cachedtiles then
		local yea = {}
		local limit = 0
		for i,v in ipairs(cachedtiles) do
			if limit >= cachepertic then break end
			table.remove(cachedtiles, i)
			table.insert(yea, v)
			limit = $ + 1
		end
		for i,v in ipairs(yea) do
			if not IsPozValid(v[1],v[2]) then continue end
			for ofer = 1,9 do
				local xoffer = (ofer%3)-1
				local yoffer = ((ofer/3)%3)-1
				if yoffer == 0 and xoffer == 0 then continue end
				if savedtiles[GetTilePos(v[1]+xoffer,v[2]+yoffer)] > 0 then continue end
				if CheckNeighbors(v[1]+xoffer,v[2]+yoffer) == 1 and IsPozValid(v[1]+xoffer,v[2]+yoffer) then
					table.insert(cachedtiles, {v[1]+xoffer,v[2]+yoffer})
				end
			end
		end
	end
end)

addHook("PostThinkFrame", function()
	if displayplayer and displayplayer.mo then
		local player = displayplayer
		local x = FixedInt(FixedFloor(player.mo.x/tilescale))
		local y = FixedInt(FixedFloor(player.mo.y/tilescale))
		local indpos = GetTilePos(x,y)
		local ddist = (renderdist*2)+1
		for i = 1,ddist^2 do
			local w = (i%ddist)-renderdist
			local j = ((i/ddist)%ddist)-renderdist
			local lol = GetTilePos(x+w,y+j)
			if savedtiles[lol] == 1 then continue end
			if not IsPozValid(x+w, y+j) then continue end 
			local icon = P_SpawnMobj(((x+w)*tilescale*FRACUNIT)+((tilescale/2)*FRACUNIT), ((y+j)*tilescale*FRACUNIT)+((tilescale/2)*FRACUNIT), P_FloorzAtPos(((x+w)*tilescale*FRACUNIT)+((tilescale/2)*FRACUNIT), ((y+j)*tilescale*FRACUNIT)+((tilescale/2)*FRACUNIT), player.mo.z, player.mo.height), MT_THOK)
			icon.radius = FRACUNIT
			icon.flags = 0
			P_ZMovement(icon)
			icon.flags = MF_NOCLIP
			icon.sprite = SPR_GRBD
			icon.renderflags = RF_FLOORSPRITE|RF_SPLATMASK|RF_NOSPLATBILLBOARD
			icon.spriteyoffset = FRACUNIT*-12
			icon.tics = 1
			icon.fuse = 1
			icon.spritexscale = (tilescale*FRACUNIT)/26
			icon.spriteyscale = (tilescale*FRACUNIT)/26
			icon.drawonlyforplayer = player
			if icon.standingslope and not flagedtiles[lol] then
				P_CreateFloorSpriteSlope(icon)
				icon.z = $ + FRACUNIT*2
			end
			if not savedtiles[lol] or savedtiles[lol] <= -1 then
				if not flagedtiles[lol] then
					icon.frame = A
				else
					if flagedtiles[lol] == 1 then
						icon.frame = C
					else
						icon.frame = N
					end
					icon.renderflags = 0
					icon.spriteyoffset = 0
				end
			elseif savedtiles[lol] == 10 then
				icon.frame = B
			elseif savedtiles[lol] == 11 then
				icon.frame = M
			elseif savedtiles[lol] < 10 then
				icon.frame = M - savedtiles[lol]
			else
				icon.frame = savedtiles[lol]
			end
		end
	end
end)

COM_AddCommand("minechance", function(player, arg)
    arg = tonumber(arg)
    if arg == nil
        CONS_Printf(player, "changes the chance that a tile is a mine")
		CONS_Printf(player, "default 10: current "+tostring(minechance))
    else
        minechance = min(max(arg, 0), 100)
    end
end, COM_ADMIN)

COM_AddCommand("renderdist", function(player, arg)
    arg = tonumber(arg)
    if arg == nil
        CONS_Printf(player, "changes how much tiles a player can see around them")
		CONS_Printf(player, "default 5: current "+tostring(renderdist))
    else
        renderdist = max(arg, 0)
    end
end, COM_ADMIN)

COM_AddCommand("resettiles", function(player, arg)
    arg = tonumber(arg)
    if arg == nil
        CONS_Printf(player, "resets all the tile info and also can be used to rescale the tiles")
		CONS_Printf(player, "default scale 64: current scale "+tostring(tilescale))
    else
        tilescale = max(arg,1)
		savedtiles = {}
		flagedtiles = {}
		cachedtiles = {}
    end
end, COM_ADMIN)

COM_AddCommand("clearcache", function(player)
	print("Tile Cache Wiped")
	print(#cachedtiles+" Caches Cleared")
	cachedtiles = {}
end, COM_ADMIN)

COM_AddCommand("iframeabuse", function(player, arg)
    arg = tonumber(arg)
    if arg == nil
		CONS_Printf(player, "2 = no iframe abuse")
        CONS_Printf(player, "1 = semi iframe abuse")
		CONS_Printf(player, "0 = full iframe abuse")
		CONS_Printf(player, "default 1: current "+tostring(iframeabuse))
    else
        iframeabuse = min(max(arg,0),2)
    end
end, COM_ADMIN)

COM_AddCommand("cachespertic", function(player, arg)
    arg = tonumber(arg)
    if arg == nil
		CONS_Printf(player, "The amount of cached tiles that will be processed per tic")
		CONS_Printf(player, "higher values make empty auto clearing faster but cause more lag")
		CONS_Printf(player, "default 4: current "+tostring(cachepertic))
    else
        cachepertic = max(arg,0)
    end
end, COM_ADMIN)

COM_AddCommand("airacivation", function(player)
	airacivate = $ ~= true
	if airacivate then
		CONS_Printf(player, "fuck you tails")
	else
		CONS_Printf(player, "tails mains be eating today")
    end
end, COM_ADMIN)

COM_AddCommand("negativemines", function(player, x, y)
	negativemines = $ ~= true
	if negativemines then
		CONS_Printf(player, "how to confuse people 101")
	else
		CONS_Printf(player, "gg to ez")
    end
end, COM_ADMIN)

COM_AddCommand("setflag", function(player, x, y)
    x = tonumber(x)
	y = tonumber(y)
    if y == nil or x == nil then
        return
    elseif abs(y) >= 10 or abs(x) >= 10 then
		return
	else
		local px = FixedInt(FixedFloor(player.mo.x/tilescale))
		local py = FixedInt(FixedFloor(player.mo.y/tilescale))
		local poz = GetTilePos(px - y, py - x)
        if flagedtiles[poz] then
			if flagedtiles[poz] == 1 and negativemines
				flagedtiles[poz] = 2
			else
				flagedtiles[poz] = nil
			end
		else
			flagedtiles[poz] = 1
		end
    end
end)

COM_AddCommand("quicksweep", function(player, x, y)
    x = tonumber(x)
	y = tonumber(y)
    if y == nil or x == nil then
        return
    elseif abs(y) >= 10 or abs(x) >= 10 then
		return
	else
		local count = 0
		local px = FixedInt(FixedFloor(player.mo.x/tilescale))
		local py = FixedInt(FixedFloor(player.mo.y/tilescale))
        QuickSweepTiles(px - y, py - x, player.mo)
    end
end)